from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_OBJECTS = [
    ArticulationObjMetaCfg(
        name="fridge_base",
        filepath="roboverse_data/assets/rlbench/close_fridge/fridge_base/usd/fridge_base.usd",
    ),
]


@configclass
class CloseFridgeMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/close_fridge/v2"
    objects = _OBJECTS
    # TODO: add checker


@configclass
class OpenFridgeMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/open_fridge/v2"
    objects = _OBJECTS
    # TODO: add checker



from __future__ import annotations

from typing import Literal

from metasim.utils.configclass import configclass


@configclass
class RandomizationMetaCfg:
    camera: bool = False
    """Randomize camera pose"""
    light: bool = False
    """Randomize light direction, temperature, intensity"""
    ground: bool = False
    """Randomize ground"""
    reflection: bool = False
    """Randomize reflection (roughness, metallic, reflectance), attach random diffuse color to surfaces that have no material"""
    table: bool = False
    """Randomize table albedo"""
    wall: bool = False
    """Add wall and roof, randomize wall"""
    scene: bool = False
    """Randomize scene"""
    level: Literal[0, 1, 2, 3] = 0
    """Randomization level"""

    def __post_init__(self):
        assert self.level in [0, 1, 2, 3]
        if self.level >= 0:
            pass
        if self.level >= 1:
            self.table = True
            self.ground = True
            self.wall = True
        if self.level >= 2:
            self.camera = True
        if self.level >= 3:
            self.light = True
            self.reflection = True



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotUprightLensForthCamera178MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/6a63fc4af48b453c91ca2b335a4d464d/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/6a63fc4af48b453c91ca2b335a4d464d_upright_lens_forth/20240826-215939_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera179MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/6a63fc4af48b453c91ca2b335a4d464d/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/6a63fc4af48b453c91ca2b335a4d464d_upright_lens_forth/20240826-215636_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera180MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/6a63fc4af48b453c91ca2b335a4d464d/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/6a63fc4af48b453c91ca2b335a4d464d_upright_lens_forth/20240826-215850_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera181MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/6a63fc4af48b453c91ca2b335a4d464d/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/6a63fc4af48b453c91ca2b335a4d464d_upright_lens_forth/20240826-214014_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera182MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/6a63fc4af48b453c91ca2b335a4d464d/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/6a63fc4af48b453c91ca2b335a4d464d_upright_lens_forth/20240826-221146_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera183MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/6a63fc4af48b453c91ca2b335a4d464d/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/6a63fc4af48b453c91ca2b335a4d464d_upright_lens_forth/20240826-220615_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera659MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f6f8675406b7437faaf96a1f82062028/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f6f8675406b7437faaf96a1f82062028_upright_lens_forth/20240826-221228_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera660MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f6f8675406b7437faaf96a1f82062028/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f6f8675406b7437faaf96a1f82062028_upright_lens_forth/20240826-220651_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera661MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f6f8675406b7437faaf96a1f82062028/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f6f8675406b7437faaf96a1f82062028_upright_lens_forth/20240826-222232_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera662MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f6f8675406b7437faaf96a1f82062028/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f6f8675406b7437faaf96a1f82062028_upright_lens_forth/20240826-215956_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera663MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f6f8675406b7437faaf96a1f82062028/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f6f8675406b7437faaf96a1f82062028_upright_lens_forth/20240826-221131_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotUprightLensForthCamera664MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="camera",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f6f8675406b7437faaf96a1f82062028/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f6f8675406b7437faaf96a1f82062028_upright_lens_forth/20240826-211423_no_interaction/trajectory-unified_wo_traj_v2.pkl"



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotHatSidewaysHat699MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="hat",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/dc02f9c1fd01432fadd6e7d15851dc37/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/dc02f9c1fd01432fadd6e7d15851dc37_hat_sideways/20240826-213912_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotHatSidewaysHat700MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="hat",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/dc02f9c1fd01432fadd6e7d15851dc37/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/dc02f9c1fd01432fadd6e7d15851dc37_hat_sideways/20240826-214945_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotHatSidewaysHat701MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="hat",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/dc02f9c1fd01432fadd6e7d15851dc37/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/dc02f9c1fd01432fadd6e7d15851dc37_hat_sideways/20240826-213225_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotHatSidewaysHat702MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="hat",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/dc02f9c1fd01432fadd6e7d15851dc37/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/dc02f9c1fd01432fadd6e7d15851dc37_hat_sideways/20240826-213412_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotHatSidewaysHat703MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="hat",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/dc02f9c1fd01432fadd6e7d15851dc37/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/dc02f9c1fd01432fadd6e7d15851dc37_hat_sideways/20240826-212424_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotHatSidewaysHat704MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="hat",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/dc02f9c1fd01432fadd6e7d15851dc37/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/dc02f9c1fd01432fadd6e7d15851dc37_hat_sideways/20240826-213224_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotHatSidewaysHat705MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="hat",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/dc02f9c1fd01432fadd6e7d15851dc37/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/dc02f9c1fd01432fadd6e7d15851dc37_hat_sideways/20240826-211626_no_interaction/trajectory-unified_wo_traj_v2.pkl"



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_OBJECTS = [
    RigidObjMetaCfg(
        name="chopping_board_visual",
        filepath="roboverse_data/assets/rlbench/put_knife_in_knife_block/chopping_board_visual/usd/chopping_board_visual.usd",
        physics=PhysicStateType.RIGIDBODY,
    ),
    RigidObjMetaCfg(
        name="knife_block_visual",
        filepath="roboverse_data/assets/rlbench/put_knife_in_knife_block/knife_block_visual/usd/knife_block_visual.usd",
        physics=PhysicStateType.GEOM,
    ),
    RigidObjMetaCfg(
        name="knife_visual",
        filepath="roboverse_data/assets/rlbench/put_knife_in_knife_block/knife_visual/usd/knife_visual.usd",
        physics=PhysicStateType.RIGIDBODY,
    ),
]


@configclass
class PutKnifeInKnifeBlockMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/put_knife_in_knife_block/v2"
    objects = _OBJECTS
    # TODO: add checker


@configclass
class PutKnifeOnChoppingBoardMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/put_knife_on_chopping_board/v2"
    objects = _OBJECTS
    # TODO: add checker



from __future__ import annotations

from abc import abstractmethod
from typing import Generic, TypeVar

import numpy as np
import torch
from loguru import logger as log

from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.types import Action, Dof, EnvState, Extra, Obs, Reward, Success, Termination, TimeOut


class BaseSimHandler:
    def __init__(self, scenario: ScenarioMetaCfg, num_envs: int = 1, headless: bool = False):
        self.scenario = scenario
        self._num_envs = num_envs
        self.demo_idxs: list[int] = list(range(num_envs))
        self.init_state_idxs: list[int] = list(range(num_envs))
        self.headless = headless

        ## For quick reference
        self.task = scenario.task
        self.robot = scenario.robot
        self.cameras = scenario.cameras
        self.objects = scenario.task.objects
        self.object_dict = {obj.name: obj for obj in self.objects}

        # For reference robot
        self.object_dict[self.robot.name] = self.robot

    def launch(self) -> None:
        pass

    ############################################################
    ## Gymnasium main methods
    ############################################################
    @abstractmethod
    def step(self, action: list[Action]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
        pass

    @abstractmethod
    def reset(self, env_ids: list[int] | None = None) -> tuple[Obs, Extra]:
        """
        Reset the environment.

        Args:
            env_ids: The indices of the environments to reset. If None, all environments are reset.

        Return:
            obs: The observation of the environment. Currently all the environments are returned. Do we need to return only the reset environments?
            extra: Extra information. Currently is empty.
        """
        pass

    @abstractmethod
    def render(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    ############################################################
    ## Set states
    ############################################################
    @abstractmethod
    def set_states(self, states: list[EnvState], env_ids: list[int] | None = None) -> None:
        pass

    @abstractmethod
    def set_dof_targets(self, obj_name: str, target: torch.Tensor) -> None:
        pass

    def set_pose(self, obj_name: str, pos: torch.Tensor, rot: torch.Tensor, env_ids: list[int] | None = None) -> None:
        states = [
            {
                obj_name: {
                    "pos": pos[env_id],
                    "rot": rot[env_id],
                }
            }
            for env_id in range(self.num_envs)
        ]
        self.set_states(states, env_ids=env_ids)

    ############################################################
    ## Get states
    ############################################################
    @abstractmethod
    def get_states(self, env_ids: list[int] | None = None) -> list[EnvState]:
        ## TODO: Add caching mechanism to speed up
        pass

    # Push from boshi, don't know if should delete
    #     def get_pos(self, obj_name: str, obj_subpath: str | None = None) -> torch.FloatTensor:
    #         return torch.stack([self.get_states()[env_id][obj_name]["pos"] for env_id in range(self.num_envs)])

    #     def get_rot(self, obj_name: str) -> torch.FloatTensor:
    #         return torch.stack([self.get_states()[env_id][obj_name]["rot"] for env_id in range(self.num_envs)])

    #     def get_dof_pos(self, obj_name: str, joint_name: str) -> torch.FloatTensor:
    #         states = self.get_states()
    #         return torch.Tensor([states[env_id][obj_name]["dof_pos"][joint_name] for env_id in range(self.num_envs)])

    def get_pos(
        self, obj_name: str, obj_subpath: str | None = None, env_ids: list[int] | None = None
    ) -> torch.FloatTensor:
        if self.num_envs > 1:
            log.warning(
                "You are using the unoptimized get_pos method, which could be slow, please contact the maintainer to"
                " support the optimized version if necessary"
            )
        if env_ids is None:
            env_ids = list(range(self.num_envs))

        states = self.get_states(env_ids=env_ids)
        return torch.stack([env_state[obj_name]["pos"] for env_state in states])

    def get_rot(self, obj_name: str, env_ids: list[int] | None = None) -> torch.FloatTensor:
        if self.num_envs > 1:
            log.warning(
                "You are using the unoptimized get_rot method, which could be slow, please contact the maintainer to"
                " support the optimized version if necessary"
            )
        if env_ids is None:
            env_ids = list(range(self.num_envs))

        states = self.get_states(env_ids=env_ids)
        return torch.stack([env_state[obj_name]["rot"] for env_state in states])

    def get_dof_pos(self, obj_name: str, joint_name: str, env_ids: list[int] | None = None) -> torch.FloatTensor:
        if self.num_envs > 1:
            log.warning(
                "You are using the unoptimized get_dof_pos method, which could be slow, please contact the maintainer"
                " to support the optimized version if necessary"
            )
        if env_ids is None:
            env_ids = list(range(self.num_envs))

        states = self.get_states(env_ids=env_ids)
        if isinstance(states[0][obj_name]["dof_pos"][joint_name], torch.Tensor):
            return torch.stack([env_state[obj_name]["dof_pos"][joint_name] for env_state in states])
        else:
            return torch.Tensor([env_state[obj_name]["dof_pos"][joint_name] for env_state in states])

    ############################################################
    ## Simulate
    ############################################################
    @abstractmethod
    def simulate(self):
        pass

    ############################################################
    ## Utils
    ############################################################
    @abstractmethod
    def refresh_render(self) -> None:
        pass

    @abstractmethod
    def get_observation(self) -> Obs:
        pass

    @abstractmethod
    def get_reward(self) -> Reward:
        pass

    @abstractmethod
    def get_success(self) -> Success:
        pass

    @abstractmethod
    def get_time_out(self) -> TimeOut:
        pass

    @abstractmethod
    def get_termination(self) -> Termination:
        pass

    ############################################################
    ## Misc
    ############################################################
    def is_running(self) -> bool:
        pass

    def get_object_joint_names(self, object: ArticulationObjMetaCfg) -> list[str]:
        assert isinstance(object, ArticulationObjMetaCfg)
        pass

    def get_robot_joint_limits(self) -> tuple[Dof, Dof]:
        pass

    @property
    def num_envs(self) -> int:
        return self._num_envs

    @property
    def episode_length_buf(self) -> list[int]:
        """
        The timestep of each environment, restart from 0 when reset, plus 1 at each step.
        """
        raise NotImplementedError


#     Push from boshi, don't know if should delete
#     def reset(self) -> tuple[Obs, Extra]:
#         obs = self.handler.get_states()
#         return obs, None

#     def step(self, actions: Action) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
#         self.handler.set_dof_targets(self.handler.robot.name, actions)
#         self.handler.simulate()
#         obs = self.handler.get_states()
#         return (
#             obs,
#             None,
#             self.handler.scenario.task.checker.check(self.handler),
#             np.zeros(self.handler.num_envs),
#             np.zeros(self.handler.num_envs),
#         )

THandler = TypeVar("THandler", bound=BaseSimHandler)


class EnvWrapper(Generic[THandler]):
    handler: THandler

    def __init__(self, *args, **kwargs) -> None: ...
    def reset(self, states: list[EnvState] | None = None) -> tuple[Obs, Extra]: ...
    def step(self, action: list[Action]) -> tuple[Obs, Reward, Success, TimeOut, Extra]: ...
    def render(self) -> None: ...
    def close(self) -> None: ...


def IdentityEnvWrapper(cls: type[THandler]) -> type[EnvWrapper[THandler]]:
    class IdentityEnv(EnvWrapper[THandler]):
        def __init__(self, *args, **kwargs):
            self.handler = cls(*args, **kwargs)
            self.handler.launch()

        def reset(self, states: list[EnvState] | None = None) -> tuple[Obs, Extra]:
            if states is not None:
                self.handler.set_states(states)
            return self.handler.reset()

        def step(self, action: list[Action]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
            return self.handler.step(action)

        def render(self) -> None:
            self.handler.render()

        def close(self) -> None:
            self.handler.close()

    return IdentityEnv


def GymEnvWrapper(cls: type[THandler]) -> type[EnvWrapper[THandler]]:
    class GymEnv:
        def __init__(self, *args, **kwargs):
            self.handler = cls(*args, **kwargs)
            self.handler.launch()

        def observation(self, states: list[EnvState]) -> Obs:
            states_v2 = states
            cameras = states_v2["cameras"].values()
            first_camera = next(iter(cameras))
            obs_rgb = torch.from_numpy(
                np.array([first_camera["rgb"][..., :3] for env_id in range(self.handler.num_envs)])
            )
            obs = {"rgb": obs_rgb}

            return obs

        def reset(self, states: list[EnvState] | None = None) -> tuple[Obs, Extra]:
            if states is not None:
                self.handler.set_states(states)
                self.handler.simulate()
            ## TODO: obs should follow unified format
            # obs = {"states": self.handler.get_states()}
            obs = self.observation(self.handler.get_states())
            return obs, None

        def step(self, actions: list[Action]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
            self.handler.set_dof_targets(self.handler.robot.name, actions)
            self.handler.simulate()

            states_v2 = self.handler.get_states()
            obs = self.observation(states_v2)
            ## TODO: implement reward, success, time_out
            return (
                obs,
                None,
                torch.zeros(self.handler.num_envs, dtype=torch.bool),
                torch.zeros(self.handler.num_envs, dtype=torch.bool),
                None,
            )

        def render(self) -> None:
            ## TODO: implement render
            pass

        def close(self) -> None:
            self.handler.close()

    return GymEnv



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotEarpieceFarHeadphone16MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/ea91786392284f03809c37976da090bf/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/ea91786392284f03809c37976da090bf_earpiece_far/20240826-214240_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone17MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/ea91786392284f03809c37976da090bf/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/ea91786392284f03809c37976da090bf_earpiece_far/20240826-212808_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone18MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/ea91786392284f03809c37976da090bf/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/ea91786392284f03809c37976da090bf_earpiece_far/20240826-215753_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone19MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/ea91786392284f03809c37976da090bf/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/ea91786392284f03809c37976da090bf_earpiece_far/20240826-220732_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone20MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/ea91786392284f03809c37976da090bf/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/ea91786392284f03809c37976da090bf_earpiece_far/20240826-220353_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone617MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/efb1727dee214d49b88c58792e4bdffc/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/efb1727dee214d49b88c58792e4bdffc_earpiece_far/20240826-212514_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone618MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/efb1727dee214d49b88c58792e4bdffc/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/efb1727dee214d49b88c58792e4bdffc_earpiece_far/20240826-212757_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone619MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/efb1727dee214d49b88c58792e4bdffc/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/efb1727dee214d49b88c58792e4bdffc_earpiece_far/20240826-213152_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone620MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/efb1727dee214d49b88c58792e4bdffc/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/efb1727dee214d49b88c58792e4bdffc_earpiece_far/20240826-215515_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone621MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/efb1727dee214d49b88c58792e4bdffc/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/efb1727dee214d49b88c58792e4bdffc_earpiece_far/20240826-213619_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone622MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/efb1727dee214d49b88c58792e4bdffc/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/efb1727dee214d49b88c58792e4bdffc_earpiece_far/20240826-213428_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone706MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f8891134871e48ce8cb5053c4287272b/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f8891134871e48ce8cb5053c4287272b_earpiece_far/20240826-213049_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone707MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f8891134871e48ce8cb5053c4287272b/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f8891134871e48ce8cb5053c4287272b_earpiece_far/20240826-211910_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone708MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f8891134871e48ce8cb5053c4287272b/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f8891134871e48ce8cb5053c4287272b_earpiece_far/20240826-215132_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone709MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f8891134871e48ce8cb5053c4287272b/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f8891134871e48ce8cb5053c4287272b_earpiece_far/20240826-214924_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone710MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f8891134871e48ce8cb5053c4287272b/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f8891134871e48ce8cb5053c4287272b_earpiece_far/20240826-212305_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotEarpieceFarHeadphone711MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="headphone",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/f8891134871e48ce8cb5053c4287272b/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/f8891134871e48ce8cb5053c4287272b_earpiece_far/20240826-213649_no_interaction/trajectory-unified_wo_traj_v2.pkl"



from metasim.cfg.objects import ArticulationObjMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PutMoneyInSafeMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/put_money_in_safe/v2"
    objects = [
        RigidObjMetaCfg(
            name="dollar_stack",
            filepath="roboverse_data/assets/rlbench/put_money_in_safe/dollar_stack/usd/dollar_stack.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        ArticulationObjMetaCfg(
            name="safe_body",
            filepath="roboverse_data/assets/rlbench/put_money_in_safe/safe_body/usd/safe_body.usd",
        ),
    ]
    # TODO: add checker



from metasim.cfg.checkers import JointPosChecker
from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.constants import BenchmarkType, TaskType
from metasim.utils import configclass

from .base_task_metacfg import BaseTaskMetaCfg


@configclass
class DrawerOpenMetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.METAWORLD
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 500
    objects = [
        ArticulationObjMetaCfg(
            name="drawer",
            filepath="data_isaaclab/assets/metaworld/drawer_open/drawer.usd",
            mjcf_path="data_isaaclab/assets_mujoco/metaworld/drawer.xml",
            fix_base_link=True,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/metaworld/rigid_body/drawer_open_v2.pkl"

    checker = JointPosChecker(
        obj_name="drawer",
        joint_name="goal_slidey",
        mode="le",
        radian_threshold=-0.16,
    )



from __future__ import annotations

import pickle as pkl

import numpy as np
import torch

from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.cfg.robots.base_robot_metacfg import BaseRobotMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.sim.base import BaseSimHandler


def get_traj_v1(task: BaseTaskMetaCfg, robot: BaseRobotMetaCfg, handler: BaseSimHandler):
    assert type(handler).__name__ == "IsaaclabHandler", "Only IsaaclabHandler is supported for v1 trajectory reading"
    demo_original = pkl.load(open(task.traj_filepath, "rb"))["demos"][robot.name]
    init_states = get_init_states(demo_original, len(demo_original), task, handler, robot)
    all_actions = get_actions(demo_original, robot, handler)
    all_states = get_all_states(demo_original, task, handler, robot)
    return init_states, all_actions, all_states


def get_actions(demo_original, robot: BaseRobotMetaCfg, handler: BaseSimHandler):
    all_actions = []
    for demo_idx in range(len(demo_original)):
        actions = []
        for action_idx in range(demo_original[demo_idx]["robot_traj"]["q"].shape[0]):
            dof_pos_target = {}
            dof_array = demo_original[demo_idx]["robot_traj"]["q"][action_idx]
            for i, joint_name in enumerate(handler.get_object_joint_names(robot)):
                value = dof_array[i]
                if type(value) is np.ndarray:
                    value = value.item()
                elif type(value) is np.float32:
                    value = float(value)
                dof_pos_target[joint_name] = value
            action = {
                "dof_pos_target": dof_pos_target,
            }
            actions.append(action)
        all_actions.append(actions)
    return all_actions


def get_init_states(demo_original: list[dict], num_envs: int, task, handler, robot):
    dds = []
    for i in range(num_envs):
        dd = {}
        ## Objects
        for obj in task.objects:
            d = {}
            d["pos"] = demo_original[i]["env_setup"][f"init_{obj.name}_pos"].copy()
            d["rot"] = demo_original[i]["env_setup"][f"init_{obj.name}_quat"]
            dof = {}
            if isinstance(obj, ArticulationObjMetaCfg):
                for joint_name in handler.get_object_joint_names(obj):
                    value = demo_original[i]["env_setup"][f"init_{joint_name}_q"]
                    if type(value) is np.ndarray:
                        value = value.item()
                    elif type(value) is np.float32:
                        value = float(value)
                    dof[joint_name] = value
                d["dof_pos"] = dof
            dd[obj.name] = d

        ## Robot
        d = {}
        d["pos"] = demo_original[i]["env_setup"]["init_robot_pos"]
        d["rot"] = demo_original[i]["env_setup"]["init_robot_quat"]
        dof = {}
        dof_array = demo_original[i]["env_setup"]["init_q"]
        for i, joint_name in enumerate(handler.get_object_joint_names(robot)):
            value = dof_array[i]
            if type(value) is np.ndarray:
                value = value.item()
            elif type(value) is np.float32:
                value = float(value)
            dof[joint_name] = value
        d["dof_pos"] = dof
        dd[robot.name] = d

        for k, d in dd.items():
            for kk in ["pos", "rot"]:
                if type(d[kk]) is np.ndarray:
                    dd[k][kk] = torch.from_numpy(d[kk])

        dds.append(dd)
    return dds


def get_all_states(demo_original, task: BaseTaskMetaCfg, handler: BaseSimHandler, robot: BaseRobotMetaCfg):
    if demo_original[0].get("object_states", None) is None:
        return None

    all_states = []
    for demo_idx in range(len(demo_original)):
        states = []
        for i in range(len(demo_original[demo_idx]["object_states"])):
            state = {}

            ## Objects
            for obj in task.objects:
                obj_state = {}
                obj_state["pos"] = demo_original[demo_idx]["object_states"][i][f"{obj.name}_pos"].copy()
                obj_state["rot"] = demo_original[demo_idx]["object_states"][i][f"{obj.name}_quat"]

                if isinstance(obj, ArticulationObjMetaCfg):
                    dof = {}
                    for joint_name in handler.get_object_joint_names(obj):
                        value = demo_original[demo_idx]["object_states"][i][f"{joint_name}_q"]
                        if type(value) is np.ndarray:
                            value = value.item()
                        elif type(value) is np.float32 or type(value) is np.float64:
                            value = float(value)
                        dof[joint_name] = value
                    obj_state["dof_pos"] = dof
                state[obj.name] = obj_state

            ## Robot
            robot_state = {}
            robot_state["pos"] = demo_original[demo_idx]["env_setup"]["init_robot_pos"]
            robot_state["rot"] = demo_original[demo_idx]["env_setup"]["init_robot_quat"]
            dof = {}
            for j, joint_name in enumerate(handler.get_object_joint_names(robot)):
                value = demo_original[demo_idx]["robot_traj"]["q"][i][j]
                if type(value) is np.ndarray:
                    value = value.item()
                elif type(value) is np.float32 or type(value) is np.float64:
                    value = float(value)
                dof[joint_name] = value
            robot_state["dof_pos"] = dof
            state[robot.name] = robot_state

            for k, obj_state in state.items():
                for kk in ["pos", "rot"]:
                    if type(obj_state[kk]) is np.ndarray:
                        state[k][kk] = torch.from_numpy(obj_state[kk])
            states.append(state)

        all_states.append(states)
    return all_states



from __future__ import annotations

import mujoco_viewer
import numpy as np
import torch
from dm_control import mjcf
from loguru import logger as log

from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import TaskType
from metasim.types import Extra, Obs, Reward, Success, Termination, TimeOut

from ..base import BaseSimHandler, EnvWrapper, IdentityEnvWrapper


class MujocoHandler(BaseSimHandler):
    def __init__(self, scenario: ScenarioMetaCfg, num_envs: int = 1, headless: bool = False):
        super().__init__(scenario, num_envs, headless)
        self.headless = headless

        if num_envs > 1:
            raise ValueError("MujocoHandler only supports single envs, please run with --num_envs 1.")

        self._task: BaseTaskMetaCfg = scenario.task
        self._robot = scenario.robot
        self._mujoco_robot_name = None
        self._robot_num_dof = None
        self._robot_path = self._robot.mjcf_path
        self._gravity_compensation = not self._robot.enabled_gravity

        self.viewer = None
        self.cameras = []
        for camera in scenario.cameras:
            self.cameras.append(camera)
        self._episode_length_buf = 0

        self.replay_traj = False
        self.use_task_decimation = False

        # FIXME: hard code decimation for now
        if self.use_task_decimation:
            self.decimation = self._task.decimation
        elif self.replay_traj:
            log.warning("Warning: hard coding decimation to 1 for object states")
            self.decimation = 1
        elif self._task.task_type == TaskType.LOCOMOTION:
            self.decimation = self._task.decimation
        else:
            log.warning("Warning: hard coding decimation to 25 for replaying trajectories")
            self.decimation = 25

    def launch(self) -> None:
        super().launch()
        model = self._init_mujoco()
        self.physics = mjcf.Physics.from_mjcf_model(model)
        self.data = self.physics.data

    def _init_mujoco(self) -> mjcf.RootElement:
        object_paths = []
        for obj in self._task.objects:
            object_paths.append(obj.mjcf_path)

        mjcf_model = mjcf.RootElement()

        ## Optional: Add ground grid
        # mjcf_model.asset.add('texture', name="texplane", type="2d", builtin="checker", width=512, height=512, rgb1=[0.2, 0.3, 0.4], rgb2=[0.1, 0.2, 0.3])
        # mjcf_model.asset.add('material', name="matplane", reflectance="0.", texture="texplane", texrepeat=[1, 1], texuniform=True)

        for camera in self.cameras:
            direction = np.array([
                camera.look_at[0] - camera.pos[0],
                camera.look_at[1] - camera.pos[1],
                camera.look_at[2] - camera.pos[2],
            ])
            direction = direction / np.linalg.norm(direction)
            up = np.array([0, 0, 1])
            right = np.cross(direction, up)
            right = right / np.linalg.norm(right)
            up = np.cross(right, direction)

            camera_params = {
                "pos": f"{camera.pos[0]} {camera.pos[1]} {camera.pos[2]}",
                "mode": "fixed",
                "fovy": camera.vertical_fov,
                "xyaxes": f"{right[0]} {right[1]} {right[2]} {up[0]} {up[1]} {up[2]}",
                "resolution": f"{camera.width} {camera.height}",
            }
            mjcf_model.worldbody.add("camera", name=f"{camera.name}_custom", **camera_params)

        # Add ground grid, light, and skybox
        mjcf_model.asset.add(
            "texture",
            name="texplane",
            type="2d",
            builtin="checker",
            width=512,
            height=512,
            rgb1=[0, 0, 0],
            rgb2=[1.0, 1.0, 1.0],
        )
        mjcf_model.asset.add(
            "material", name="matplane", reflectance="0.2", texture="texplane", texrepeat=[1, 1], texuniform=True
        )
        ground = mjcf_model.worldbody.add(
            "geom",
            type="plane",
            pos="0 0 0",
            size="100 100 0.001",
            quat="1 0 0 0",
            condim="3",
            conaffinity="15",
            material="matplane",
        )
        self.object_body_names = []
        self.mj_objects = {}
        for i, (obj, obj_path) in enumerate(zip(self._task.objects, object_paths)):
            obj_mjcf = mjcf.from_path(obj_path)
            obj_attached = mjcf_model.attach(obj_mjcf)
            obj_attached.add("freejoint")
            self.object_body_names.append(obj_attached.full_identifier)
            self.mj_objects[obj.name] = obj_mjcf

        robot_xml = mjcf.from_path(self._robot_path)
        robot_attached = mjcf_model.attach(robot_xml)
        if self._robot.freejoint:
            robot_attached.add("freejoint")
        self.robot_attached = robot_attached
        self.mj_objects[self._robot.name] = robot_xml
        self._mujoco_robot_name = robot_xml.full_identifier
        return mjcf_model

    def _get_root_state(self, obj, obj_name):
        """Get root position, rotation, velocity for an object or robot."""
        if obj == self._robot:
            if self._robot.freejoint:
                root_joint = self.physics.data.joint(self._mujoco_robot_name)
                return {
                    "pos": torch.tensor(root_joint.qpos[:3], dtype=torch.float32),
                    "rot": torch.tensor(root_joint.qpos[3:7], dtype=torch.float32),
                    "vel": torch.tensor(root_joint.qvel[:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(root_joint.qvel[3:6], dtype=torch.float32),
                }
            else:
                root_body_id = self.physics.model.body(self._mujoco_robot_name).id
                return {
                    "pos": torch.tensor(self.physics.data.xpos[root_body_id], dtype=torch.float32),
                    "rot": torch.tensor(self.physics.data.xquat[root_body_id], dtype=torch.float32),
                    "vel": torch.tensor(self.physics.data.cvel[root_body_id][:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(self.physics.data.cvel[root_body_id][3:6], dtype=torch.float32),
                }
        else:
            model_name = self.mj_objects[obj_name].model + "/"
            try:
                obj_joint = self.physics.data.joint(model_name)
                return {
                    "pos": torch.tensor(obj_joint.qpos[:3], dtype=torch.float32),
                    "rot": torch.tensor(obj_joint.qpos[3:7], dtype=torch.float32),
                    "vel": torch.tensor(obj_joint.qvel[:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(obj_joint.qvel[3:6], dtype=torch.float32),
                }
            except KeyError:
                obj_body_id = self.physics.model.body(model_name).id
                return {
                    "pos": torch.tensor(self.physics.data.xpos[obj_body_id], dtype=torch.float32),
                    "rot": torch.tensor(self.physics.data.xquat[obj_body_id], dtype=torch.float32),
                    "vel": torch.tensor(self.physics.data.cvel[obj_body_id][:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(self.physics.data.cvel[obj_body_id][3:6], dtype=torch.float32),
                }

    def _get_joint_states(self, obj_name):
        """Get joint positions and velocities."""
        joint_states = {"dof_pos": {}, "dof_vel": {}}

        for joint_id in range(self.physics.model.njnt):
            joint_name = self.physics.model.joint(joint_id).name
            if joint_name.startswith(f"{obj_name}/") or (
                obj_name == self._robot.name and joint_name.startswith(self._mujoco_robot_name)
            ):
                clean_joint_name = joint_name.split("/")[-1]
                if clean_joint_name == "":
                    continue
                joint = self.physics.data.joint(joint_name)

                if len(joint.qpos.shape) == 0 or joint.qpos.shape == (1,):
                    joint_states["dof_pos"][clean_joint_name] = joint.qpos.item()
                    joint_states["dof_vel"][clean_joint_name] = joint.qvel.item()
                else:  # free joint
                    joint_states["dof_pos"][clean_joint_name] = torch.tensor(joint.qpos.copy(), dtype=torch.float32)
                    joint_states["dof_vel"][clean_joint_name] = torch.tensor(joint.qvel.copy(), dtype=torch.float32)

        return joint_states

    def _get_actuator_states(self, obj_name):
        """Get actuator states (targets and forces)."""
        actuator_states = {
            "dof_pos_target": {},
            "dof_vel_target": {},
            "dof_torque": {},
        }

        for actuator_id in range(self.physics.model.nu):
            actuator = self.physics.model.actuator(actuator_id)
            if actuator.name.startswith(self._mujoco_robot_name):
                clean_name = actuator.name[len(self._mujoco_robot_name) :]

                actuator_states["dof_pos_target"][clean_name] = float(
                    self.physics.data.ctrl[actuator_id].item()
                )  # Hardcoded to position control
                actuator_states["dof_vel_target"][clean_name] = None
                actuator_states["dof_torque"][clean_name] = float(self.physics.data.actuator_force[actuator_id].item())

        return actuator_states

    def get_states(self, env_ids: list[int] | None = None) -> list[dict]:
        state = {}

        for obj in self._task.objects + [self._robot]:
            state[obj.name] = self._get_root_state(obj, obj.name)
            state[obj.name].update(self._get_joint_states(obj.name))
            if obj == self._robot:
                state[obj.name].update(self._get_actuator_states(obj.name))

        for body_id in range(self.physics.model.nbody):
            body = self.physics.model.body(body_id)
            if body.name not in [self._mujoco_robot_name] + [obj.name for obj in self._task.objects]:
                state[f"metasim_body_{body.name}"] = {
                    "pos": torch.tensor(self.physics.data.xpos[body_id], dtype=torch.float32),
                    "rot": torch.tensor(self.physics.data.xquat[body_id], dtype=torch.float32),
                    "vel": torch.tensor(self.physics.data.cvel[body_id][:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(self.physics.data.cvel[body_id][3:6], dtype=torch.float32),
                    "com": torch.tensor(self.physics.data.subtree_com[body_id], dtype=torch.float32),
                    "com_vel": torch.tensor(self.physics.data.subtree_linvel[body_id], dtype=torch.float32),
                }

        for site_id in range(self.physics.model.nsite):
            site = self.physics.model.site(site_id)

            linear_vel = None
            angular_vel = None
            for sensor_id in range(self.physics.model.nsensor):
                sensor = self.physics.model.sensor(sensor_id)
                if self.physics.model.sensor_objid[sensor_id] == site_id:
                    # mjtSensor enum: VELOCIMETER=2, GYRO=3
                    if sensor.type == 2:
                        linear_vel = torch.tensor(
                            self.physics.data.sensordata[sensor_id : sensor_id + 3], dtype=torch.float32
                        )
                    elif sensor.type == 3:
                        angular_vel = torch.tensor(
                            self.physics.data.sensordata[sensor_id : sensor_id + 3], dtype=torch.float32
                        )

            state[f"metasim_site_{site.name}"] = {
                "pos": torch.tensor(self.physics.data.site_xpos[site_id], dtype=torch.float32),
                "rot": torch.tensor(self.physics.data.site_xmat[site_id], dtype=torch.float32),
                "vel": linear_vel if linear_vel is not None else None,
                "ang_vel": angular_vel if angular_vel is not None else None,
            }

        return [state]

    def _set_root_state(self, obj_name, obj_state, zero_vel=False):
        """Set root position and rotation."""
        if "pos" not in obj_state and "rot" not in obj_state:
            return

        if obj_name == self._robot.name:
            if self._robot.freejoint:
                root_joint = self.physics.data.joint(self._mujoco_robot_name)
                root_joint.qpos[:3] = obj_state.get("pos", [0, 0, 0])
                root_joint.qpos[3:7] = obj_state.get("rot", [1, 0, 0, 0])
                if zero_vel:
                    root_joint.qvel[:6] = 0
            else:
                root_body = self.physics.named.model.body_pos[self._mujoco_robot_name]
                root_body_quat = self.physics.named.model.body_quat[self._mujoco_robot_name]
                root_body[:] = obj_state.get("pos", [0, 0, 0])
                root_body_quat[:] = obj_state.get("rot", [1, 0, 0, 0])
        else:
            model_name = self.mj_objects[obj_name].model + "/"
            try:
                obj_joint = self.physics.data.joint(model_name)
                obj_joint.qpos[:3] = obj_state["pos"]
                obj_joint.qpos[3:7] = obj_state["rot"]
                if zero_vel:
                    obj_joint.qvel[:6] = 0
            except KeyError:
                obj_body = self.physics.named.model.body_pos[model_name]
                obj_body_quat = self.physics.named.model.body_quat[model_name]
                obj_body[:] = obj_state["pos"]
                obj_body_quat[:] = obj_state["rot"]

    def _set_joint_state(self, obj_name, obj_state, zero_vel=False):
        """Set joint positions."""
        if "dof_pos" not in obj_state:
            return

        for joint_name, joint_pos in obj_state["dof_pos"].items():
            full_joint_name = (
                f"{self._mujoco_robot_name}{joint_name}" if obj_name == self._robot.name else f"{obj_name}/{joint_name}"
            )
            joint = self.physics.data.joint(full_joint_name)
            joint.qpos = joint_pos
            if zero_vel:
                joint.qvel = 0
            try:
                actuator = self.physics.model.actuator(full_joint_name)
                self.physics.data.ctrl[actuator.id] = joint_pos
            except KeyError:
                pass

    def set_states(self, states, env_ids=None, zero_vel=True):
        super().set_states(states, env_ids)
        if len(states) > 1:
            raise ValueError("MujocoHandler only supports single env state setting")

        for obj_name, obj_state in states[0].items():
            self._set_root_state(obj_name, obj_state, zero_vel)
            self._set_joint_state(obj_name, obj_state, zero_vel)

        if self.viewer is not None:
            self.physics.forward()
            self.viewer.render()

    def _disable_robotgravity(self):
        gravity_vec = np.array([0.0, 0.0, -9.81])

        self.physics.data.xfrc_applied[:] = 0
        for body_name in self.robot_body_names:
            body_id = self.physics.model.body(body_name).id
            force_vec = -gravity_vec * self.physics.model.body(body_name).mass
            self.physics.data.xfrc_applied[body_id, 0:3] = force_vec
            self.physics.data.xfrc_applied[body_id, 3:6] = 0

    def step(self, action: list[float]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
        self._episode_length_buf += 1

        super().step(action)
        joint_targets = action[0]["dof_pos_target"]

        if self.replay_traj:
            for joint_name, target_pos in joint_targets.items():
                joint = self.physics.data.joint(f"{self._mujoco_robot_name}{joint_name}")
                joint.qpos = target_pos
        else:
            for joint_name, target_pos in joint_targets.items():
                actuator = self.physics.data.actuator(f"{self._mujoco_robot_name}{joint_name}")
                actuator.ctrl = target_pos

        if self._gravity_compensation:
            self._disable_robotgravity()
        self.physics.step(self.decimation)

        if self.viewer is not None and self.viewer.is_alive and not self.headless:
            self.viewer.render()

        obs = self.get_observation() if self.headless else None
        success = self.task.checker.check(self)
        reward = self.get_reward()
        return (
            obs,
            reward,
            success,
            torch.tensor([self._episode_length_buf >= self._task.episode_length]),
            {},
        )

    def reset(self, env_ids: list[int] | None = None) -> tuple[Obs, Extra]:
        assert env_ids is None or env_ids == [0], "MujocoHandler only supports single env reset"

        self._episode_length_buf = 0

        self.joint_names = [self.physics.model.joint(i).name for i in range(self.physics.model.njnt)]
        self.body_names = [self.physics.model.body(i).name for i in range(self.physics.model.nbody)]
        self.robot_body_names = [
            body_name for body_name in self.body_names if body_name.startswith(self._mujoco_robot_name)
        ]
        self.articulation_body_names = []
        for obj in self._task.objects:
            if isinstance(obj, ArticulationObjMetaCfg):
                self.articulation_body_names.append(obj.name)
        obj_joints = []
        for joint_name in self.joint_names:
            obj_joints.append(self.physics.data.joint(joint_name))
        self.obj_joints = obj_joints

        if self.viewer is not None:
            self.viewer.close()

        self.physics.forward()
        self.physics.step()
        obs = None
        if not self.headless:
            self.viewer = mujoco_viewer.MujocoViewer(self.physics.model.ptr, self.physics.data.ptr)
            self.viewer.render()
        else:
            obs = self.get_observation()

        self.task.checker.reset(self)
        return obs, {}

    ############################################################
    ## Utils
    ############################################################
    def refresh_render(self) -> None:
        pass  # XXX: Is it ok?

    def render(self):
        if self.headless:
            # Return an image in headless mode
            # TODO: config the camera in both cfg and xml
            return self.physics.render(camera_id=1, width=640, height=480)
        else:
            if self.viewer is not None and self.viewer.is_alive:
                self.viewer.render()
                # Shouldn't return image for non-headless mode, this will block the desktop render loop
                # return self.physics.render(camera_id=1, width=640, height=480)

    def get_observation(self) -> Obs:
        obs = {}
        # Shouldn't return image for non-headless mode, this will block the desktop render loop
        if len(self.cameras) == 0 or not self.headless:
            return obs
        camera = self.cameras[0]  # Using first camera as specified
        camera_id = "camera_custom"  # XXX: hard code camera id for now

        if "rgb" in camera.data_types:
            rgb_img = self.physics.render(
                width=camera.width,
                height=camera.height,
                camera_id=camera_id,
                depth=False,
            )
            obs["rgb"] = torch.from_numpy(rgb_img.copy()).unsqueeze(0)

        if "depth" in camera.data_types:
            depth_img = self.physics.render(
                width=camera.width,
                height=camera.height,
                camera_id=camera_id,
                depth=True,
            )
            obs["depth"] = torch.from_numpy(depth_img.copy()).unsqueeze(0)

        return obs

    def get_reward(self) -> Reward:
        from metasim.cfg.tasks.base_task_metacfg import BaseRLTaskMetaCfg

        if hasattr(self.task, "reward_fn"):
            return self.task.reward_fn(self.get_states())

        if isinstance(self.task, BaseRLTaskMetaCfg):
            final_reward = torch.zeros(self.num_envs)
            for reward_func, reward_weight in zip(self.task.reward_functions, self.task.reward_weights):
                final_reward += reward_func(self._robot.name)(self.get_states()) * reward_weight
            return final_reward
        else:
            return None

    def get_success(self) -> Success:
        return self.task.checker.check(self)

    def get_time_out(self) -> TimeOut:
        return torch.tensor([self._episode_length_buf >= self._task.episode_length], dtype=torch.bool)

    def get_termination(self) -> Termination:
        # TODO: implement termination in downstream tasks
        pass

    ############################################################
    ## Misc
    ############################################################
    def get_pos(
        self, obj_name: str, obj_subpath: str | None = None, env_ids: list[int] | None = None
    ) -> torch.FloatTensor:
        return torch.tensor(self.get_states()[0][obj_name]["pos"], dtype=torch.float32).unsqueeze(0)

    def is_running(self) -> bool:
        return True

    @property
    def num_envs(self) -> int:
        return 1

    @property
    def episode_length_buf(self) -> list[int]:
        return [self._episode_length_buf]


MujocoEnv: type[EnvWrapper[MujocoHandler]] = IdentityEnvWrapper(MujocoHandler)



from metasim.cfg.checkers import DetectedChecker, RelativeBboxDetector
from metasim.cfg.objects import (
    PrimitiveCubeMetaCfg,
)
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .maniskill_task_metacfg import ManiskillTaskMetaCfg


@configclass
class StackGreenCubeMetaCfg(ManiskillTaskMetaCfg):
    episode_length = 250
    objects = [
        PrimitiveCubeMetaCfg(
            name="cube",
            size=[0.04, 0.04, 0.04],
            mass=0.02,
            physics=PhysicStateType.RIGIDBODY,
            color=[0.0, 1.0, 0.0],
            mjcf_path="data_isaaclab/assets_mujoco/maniskill2/stack_cube/cube.mjcf",
        ),
        PrimitiveCubeMetaCfg(
            name="base",
            size=[0.04, 0.04, 0.04],
            mass=0.02,
            physics=PhysicStateType.RIGIDBODY,
            color=[0.0, 0.0, 1.0],
            mjcf_path="data_isaaclab/assets_mujoco/maniskill2/stack_cube/base.mjcf",
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/maniskill2/rigid_body/StackCube-v0/trajectory-unified_v2.pkl"

    ## TODO: detect velocity
    checker = DetectedChecker(
        obj_name="cube",
        detector=RelativeBboxDetector(
            base_obj_name="base",
            relative_pos=(0.0, 0.0, 0.04),
            relative_quat=(1.0, 0.0, 0.0, 0.0),
            checker_lower=(-0.02, -0.02, -0.02),
            checker_upper=(0.02, 0.02, 0.02),
            ignore_base_ori=True,
        ),
    )



from metasim.cfg.checkers import PositionShiftChecker
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, TaskType
from metasim.utils import configclass


@configclass
class PushBlueBlockRightMetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.CALVIN
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    checker = PositionShiftChecker(
        obj_name="block_blue",
        distance=0.065,
        axis="x",
    )



import os

import torch
from loguru import logger as log

from metasim.cfg.objects import (
    ArticulationObjMetaCfg,
    BaseObjMetaCfg,
    PrimitiveCubeMetaCfg,
    PrimitiveSphereMetaCfg,
    RigidObjMetaCfg,
)
from metasim.cfg.robots import BaseRobotMetaCfg
from metasim.constants import PhysicStateType

try:
    from .empty_env import EmptyEnv
except:
    pass


def add_objects(env: "EmptyEnv", objects: list[BaseObjMetaCfg]) -> None:
    import omni.isaac.lab.sim as sim_utils
    from omni.isaac.lab.assets import (
        Articulation,
        ArticulationCfg,
        RigidObject,
        RigidObjectCfg,
    )

    for obj in objects:
        assert isinstance(obj, BaseObjMetaCfg)
        prim_path = f"/World/envs/env_.*/{obj.name}"
        ## Rigid object
        if isinstance(obj, RigidObjMetaCfg):
            if obj.physics == PhysicStateType.XFORM:
                rigid_props = sim_utils.RigidBodyPropertiesCfg(disable_gravity=True, kinematic_enabled=True)
                collision_props = None
            elif obj.physics == PhysicStateType.GEOM:
                rigid_props = sim_utils.RigidBodyPropertiesCfg(disable_gravity=True, kinematic_enabled=True)
                collision_props = sim_utils.CollisionPropertiesCfg(collision_enabled=True)
            elif obj.physics == PhysicStateType.RIGIDBODY:
                rigid_props = sim_utils.RigidBodyPropertiesCfg()
                collision_props = sim_utils.CollisionPropertiesCfg(collision_enabled=True)
            else:
                raise ValueError(f"Unsupported physics state: {obj.physics}")

            ## Primitive object
            if isinstance(obj, PrimitiveCubeMetaCfg):
                env.scene.rigid_objects[obj.name] = RigidObject(
                    RigidObjectCfg(
                        prim_path=prim_path,
                        spawn=sim_utils.MeshCuboidCfg(
                            size=tuple([x * obj.scale for x in obj.size]),
                            mass_props=sim_utils.MassPropertiesCfg(mass=obj.mass),
                            visual_material=sim_utils.PreviewSurfaceCfg(
                                diffuse_color=(obj.color[0], obj.color[1], obj.color[2])
                            ),
                            rigid_props=rigid_props,
                            collision_props=collision_props,
                        ),
                    )
                )
                continue
            if isinstance(obj, PrimitiveSphereMetaCfg):
                env.scene.rigid_objects[obj.name] = RigidObject(
                    RigidObjectCfg(
                        prim_path=prim_path,
                        spawn=sim_utils.MeshSphereCfg(
                            radius=obj.radius,
                            mass_props=sim_utils.MassPropertiesCfg(mass=obj.mass),
                            visual_material=sim_utils.PreviewSurfaceCfg(
                                diffuse_color=(obj.color[0], obj.color[1], obj.color[2])
                            ),
                            rigid_props=rigid_props,
                            collision_props=collision_props,
                        ),
                    )
                )
                continue
            ## File-based object
            usd_file_cfg = sim_utils.UsdFileCfg(
                usd_path=obj.filepath,
                rigid_props=rigid_props,
                collision_props=collision_props,
                scale=(obj.scale, obj.scale, obj.scale),
            )
            if isinstance(obj, RigidObjMetaCfg):
                env.scene.rigid_objects[obj.name] = RigidObject(RigidObjectCfg(prim_path=prim_path, spawn=usd_file_cfg))
                continue

        ## Articulation object
        if isinstance(obj, ArticulationObjMetaCfg):
            env.scene.articulations[obj.name] = Articulation(
                ArticulationCfg(
                    prim_path=prim_path,
                    spawn=sim_utils.UsdFileCfg(usd_path=obj.filepath),
                    actuators={},
                )
            )
            continue
        raise ValueError(f"Unsupported object type: {type(obj)}")


def add_robot(env: "EmptyEnv", robot: BaseRobotMetaCfg) -> None:
    import omni.isaac.lab.sim as sim_utils
    from omni.isaac.lab.actuators import ImplicitActuatorCfg
    from omni.isaac.lab.assets import Articulation, ArticulationCfg

    cfg = ArticulationCfg(
        spawn=sim_utils.UsdFileCfg(
            usd_path=robot.filepath,
            rigid_props=sim_utils.RigidBodyPropertiesCfg(),
            articulation_props=sim_utils.ArticulationRootPropertiesCfg(),
        ),
        actuators={
            joint_name: ImplicitActuatorCfg(joint_names_expr=[joint_name], stiffness=None, damping=None)
            for joint_name in robot.actuators
        },
    )
    cfg.prim_path = f"/World/envs/env_.*/{robot.name}"
    cfg.spawn.usd_path = os.path.abspath(robot.filepath)
    cfg.spawn.rigid_props.disable_gravity = not robot.enabled_gravity
    cfg.spawn.articulation_props.enabled_self_collisions = robot.enabled_self_collisions
    for joint_name, actuator in robot.actuators.items():
        cfg.actuators[joint_name].velocity_limit = actuator.velocity_limit

    robot_inst = Articulation(cfg)
    env.scene.articulations[robot.name] = robot_inst
    env.robot = robot_inst


def get_pose(
    env: "EmptyEnv", obj_name: str, obj_subpath: str | None = None, env_ids: list[int] | None = None
) -> tuple[torch.FloatTensor, torch.FloatTensor]:
    from omni.isaac.core.prims import RigidPrimView

    if env_ids is None:
        env_ids = list(range(env.num_envs))

    if obj_name in env.scene.rigid_objects:
        obj_inst = env.scene.rigid_objects[obj_name]
    elif obj_name in env.scene.articulations:
        obj_inst = env.scene.articulations[obj_name]
    else:
        raise ValueError(f"Object {obj_name} not found")

    if obj_subpath is None:
        pos = obj_inst.data.root_pos_w.cpu()[env_ids] - env.scene.env_origins.cpu()[env_ids]
        rot = obj_inst.data.root_quat_w.cpu()[env_ids]
    else:
        ## XXX: I suspect some prim may failed to apply a RigidPrimView
        view = RigidPrimView(
            obj_inst.cfg.prim_path + "/" + obj_subpath,
            name=f"{obj_name}_{obj_subpath}_view",
            reset_xform_properties=False,
        )
        pos, rot = view.get_world_poses(indices=env_ids)
        pos = pos - env.scene.env_origins[env_ids]
        pos = pos.cpu()
        rot = rot.cpu()

    assert pos.shape == (len(env_ids), 3)
    assert rot.shape == (len(env_ids), 4)
    return pos, rot


def joint_is_implicit_actuator(joint_name: str, obj_inst) -> bool:
    from omni.isaac.lab.actuators import ImplicitActuatorCfg
    from omni.isaac.lab.assets import Articulation

    assert isinstance(obj_inst, Articulation)
    actuators = [actuator for actuator in obj_inst.actuators.values() if joint_name in actuator.joint_names]
    if len(actuators) == 0:
        log.error(f"Joint {joint_name} could not be found in actuators of {obj_inst.cfg.prim_path}")
        return False
    if len(actuators) > 1:
        log.warning(f"Joint {joint_name} is found in multiple actuators of {obj_inst.cfg.prim_path}")
    actuator = actuators[0]
    return isinstance(actuator, ImplicitActuatorCfg)



from __future__ import annotations

import logging
import os
import time
from dataclasses import dataclass
from typing import Literal

import rootutils
import tyro
from loguru import logger as log
from rich.logging import RichHandler

rootutils.setup_root(__file__, pythonpath=True)
from metasim.cfg.randomization import RandomizationMetaCfg

# 这行只是让 Python 识别 logging 的 TRACE 等级，可按需保留
logging.addLevelName(5, "TRACE")
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])


@dataclass
class Args:
    random: RandomizationMetaCfg
    task: str
    robot: Literal[
        "franka",
        "franka_with_gripper_extension",
        "sawyer",
        "iiwa",
        "h1_2_without_hand",
        "ur5e_2f85",
    ] = "franka"
    num_envs: int = 1
    sim: Literal["isaaclab", "isaacgym", "pyrep", "pybullet", "sapien", "mujoco"] = "isaaclab"
    render: Literal["isaaclab", "isaacgym", "pyrep", "pybullet", "sapien", "mujoco"] | None = None
    object_states: bool = False
    save_obs: bool = True


def parse_args() -> Args:
    args = tyro.cli(Args)
    return args


###########################################################
## Utils
###########################################################
def get_actions(all_actions, action_idx: int, num_envs: int):
    # 按当前步索引各个 env 的动作，如果超出序列就取最后一个动作。
    envs_actions = all_actions[:num_envs]
    actions = [
        env_actions[action_idx] if action_idx < len(env_actions) else env_actions[-1] for env_actions in envs_actions
    ]
    return actions


def get_states(all_states, action_idx: int, num_envs: int):
    # 按当前步索引各个 env 的状态，如果超出序列就取最后一个状态。
    envs_states = all_states[:num_envs]
    states = [env_states[action_idx] if action_idx < len(env_states) else env_states[-1] for env_states in envs_states]
    return states


def save_obs(obs, step: int):
    # 保存观察到的图像，可按需修改、裁剪或注释掉。
    from torchvision.utils import make_grid, save_image

    if obs is None or obs.get("rgb", None) is None:
        return

    rgb_data = obs["rgb"]  # (N, H, W, C), uint8
    # 这里假设 env 数量比较小，所以直接拼成网格
    import torch

    tensor_data = torch.from_numpy(rgb_data).permute(0, 3, 1, 2)  # (N, C, H, W)
    image = make_grid(tensor_data / 255, nrow=int(rgb_data.shape[0] ** 0.5))  # (C, H, W)
    os.makedirs("tmp", exist_ok=True)
    save_image(image, f"tmp/rgb_all_{step:04d}.png")


###########################################################
## Main
###########################################################
def main():
    args = parse_args()
    num_envs: int = args.num_envs

    # 如果使用IsaacGym，需要在import前先做gym相关的初始化
    if args.sim == "isaacgym" or args.render == "isaacgym":
        from isaacgym import gymapi, gymtorch, gymutil  # noqa: F401

    # 一些 import 放后面以支持 isaacgym
    from metasim.cfg.cameras import PinholeCameraMetaCfg
    from metasim.cfg.scenario import ScenarioMetaCfg
    from metasim.constants import SimType
    from metasim.double_sim import DoubleSimEnv
    from metasim.utils.demo_util import get_traj
    from metasim.utils.setup_util import get_robot, get_sim_env_class, get_task

    # 构造 scenario
    task = get_task(args.task)
    robot = get_robot(args.robot)
    camera = PinholeCameraMetaCfg(
        name="camera",
        data_types=["rgb", "depth"],
        width=256,
        height=256,
        pos=(1.5, -1.5, 1.5),
        look_at=(0.0, 0.0, 0.0),
    )
    scenario = ScenarioMetaCfg(task=task, robot=robot, cameras=[camera], random=args.random)

    # 创建环境：单环境或DoubleSimEnv
    if args.render is None:
        env_class = get_sim_env_class(SimType(args.sim))
        env = env_class(scenario, num_envs)
    else:
        env_class_render = get_sim_env_class(SimType(args.render))
        env_render = env_class_render(scenario, num_envs)
        env_class_physics = get_sim_env_class(SimType(args.sim))
        env_physics = env_class_physics(scenario, num_envs)
        env = DoubleSimEnv(env_physics, env_render)

    # 从 pkl 读取多段 demo
    # 假设 get_traj 返回类似：
    #   init_states: List[init_state1, init_state2, ...]
    #   all_actions: List[action_seq1, action_seq2, ...]
    #   all_states:  List[state_seq1,  state_seq2,  ...]
    # 其中下标 i 对应第 i 段 demo
    assert os.path.exists(task.traj_filepath), f"Trajectory file: {task.traj_filepath} does not exist."

    tic = time.time()
    init_states_list, all_actions_list, all_states_list = get_traj(task, robot, env.handler)
    toc = time.time()
    log.trace(f"Time to load data: {toc - tic:.2f}s")

    # 逐段回放：直到所有 demo 播放完毕才退出
    for demo_idx in range(len(init_states_list)):
        log.info(f"========== Replaying Demo #{demo_idx} ==========")

        init_states = [init_states_list[demo_idx]]  # 若你有多个env，也可对该数据做相应扩展
        actions_seq = [all_actions_list[demo_idx]]
        states_seq = [all_states_list[demo_idx]]

        # 先reset到该demo的初始状态
        tic = time.time()
        obs, extras = env.reset(states=init_states[:num_envs])
        toc = time.time()
        log.trace(f"Time to reset Demo #{demo_idx}: {toc - tic:.2f}s")

        step = 0
        if args.save_obs:
            save_obs(obs, step)

        # 连续回放单个 demo
        while True:
            log.debug(f"Demo #{demo_idx}, Step={step}")
            if args.object_states:
                # 如果要直接回放 states
                if args.sim == "isaacgym":
                    raise NotImplementedError("IsaacGym does not support object states replay yet.")
                current_states = get_states(states_seq, step, num_envs)
                env.handler.set_states(current_states)
                env.handler.refresh_render()
                obs = env.handler.get_observation()

                # 判断是否完成
                success = env.handler.task.checker.check(env.handler)
                if success.any():
                    log.info(f"Env {success.nonzero().squeeze(-1).tolist()} succeeded!")
                if success.all():
                    break
            else:
                # 正常执行 actions
                current_actions = get_actions(actions_seq, step, num_envs)
                obs, reward, success, time_out, extras = env.step(current_actions)

                if success.any():
                    log.info(f"Env {success.nonzero().squeeze(-1).tolist()} succeeded!")
                if time_out.any():
                    log.info(f"Env {time_out.nonzero().squeeze(-1).tolist()} timed out!")

                if success.all() or time_out.all():
                    break

            if args.save_obs:
                save_obs(obs, step + 1)

            step += 1

        log.info(f"Demo #{demo_idx} finished at step={step}")

    log.info("All demos in the pkl file have been replayed. Exiting.")


if __name__ == "__main__":
    main()



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PutRubbishInBinMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/put_rubbish_in_bin/v2"
    objects = [
        RigidObjMetaCfg(
            name="bin_visual",
            filepath="roboverse_data/assets/rlbench/put_rubbish_in_bin/bin_visual/usd/bin_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="tomato1_visual",
            filepath="roboverse_data/assets/rlbench/put_rubbish_in_bin/tomato1_visual/usd/tomato1_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="tomato2_visual",
            filepath="roboverse_data/assets/rlbench/put_rubbish_in_bin/tomato2_visual/usd/tomato2_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="rubbish_visual",
            filepath="roboverse_data/assets/rlbench/put_rubbish_in_bin/rubbish_visual/usd/rubbish_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from __future__ import annotations

import logging

import numpy as np
import torch
from dm_control.utils import rewards
from loguru import logger as log
from rich.logging import RichHandler

from metasim.cfg.tasks.base_task_metacfg import BaseRLTaskMetaCfg
from metasim.constants import BenchmarkType, TaskType
from metasim.types import EnvState
from metasim.utils import configclass
from metasim.utils.h1_utils import (
    actuator_forces,
    center_of_mass_velocity,
    head_height,
    torso_upright,
)

logging.addLevelName(5, "TRACE")
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])

########################################################
## Constants adapted from humanoid_bench/tasks/basic_locomotion_envs.py
########################################################

# Height of head above which stand reward is 1.
_H1_STAND_HEIGHT = 1.65
_H1_CRAWL_HEIGHT = 0.8
_G1_STAND_HEIGHT = 1.28
_G1_CRAWL_HEIGHT = 0.6


@configclass
class HumanoidTaskMetaCfg(BaseRLTaskMetaCfg):
    decimation: int = 10
    source_benchmark = BenchmarkType.HUMANOIDBENCH
    task_type = TaskType.LOCOMOTION
    episode_length = 800  # TODO: may change
    objects = []
    reward_weights = [1.0]

    @staticmethod
    def observation_space(envstates: list[EnvState]) -> torch.Tensor:
        # TODO: currently hardcoding to [-inf,inf] in gym_env_wrapper
        pass

    @staticmethod
    def observation_function(envstates: list[EnvState]) -> torch.Tensor:
        # TODO: align torch.tensor or numpy.array
        env_obs = []
        for envstate in envstates:
            flattened = []
            for obj_name, obj_state in envstate.items():
                if obj_name == "h1" or obj_name == "g1" or obj_name == "h1_simple_hand":
                    for key in ["pos", "rot", "vel", "ang_vel"]:
                        if key in obj_state and obj_state[key] is not None:
                            flattened.append(np.array(obj_state[key]).flatten())

                    if "dof_pos" in obj_state:
                        for _, value in obj_state["dof_pos"].items():
                            if isinstance(value, (np.ndarray, float)):
                                flattened.append(np.array(value).flatten())

                    if "dof_vel" in obj_state:
                        for key, value in obj_state["dof_vel"].items():
                            if isinstance(value, (np.ndarray, float)):
                                flattened.append(np.array(value).flatten())
                    continue

                elif not obj_name.startswith("metasim"):
                    # flatten pos, rot, vel, ang_vel for other object. (state-based observation)
                    for key in ["pos", "rot", "vel", "ang_vel"]:
                        if key in obj_state and obj_state[key] is not None:
                            flattened.append(obj_state[key].numpy())

                env_obs.append(np.concatenate([arr.flatten() for arr in flattened]))

        return env_obs


class BaseLocomotionReward:
    _move_speed = None
    htarget_low = np.array([-1.0, -1.0, 0.8])
    htarget_high = np.array([1000.0, 1.0, 2.0])
    success_bar = None

    def __init__(self, robot_name="h1"):
        self.robot_name = robot_name
        if robot_name == "h1" or robot_name == "h1_simple_hand":
            self._stand_height = _H1_STAND_HEIGHT
            self._crawl_height = _H1_CRAWL_HEIGHT
        elif robot_name == "g1":
            self._stand_height = _G1_STAND_HEIGHT
            self._crawl_height = _G1_CRAWL_HEIGHT
        else:
            raise ValueError(f"Unknown robot {robot_name}")

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        ret_rewards = []
        ret_infos = []
        for state in states:
            standing = rewards.tolerance(
                head_height(state, self.robot_name),
                bounds=(self._stand_height, float("inf")),
                margin=self._stand_height / 4,
            )
            upright = rewards.tolerance(
                torso_upright(state, self.robot_name),
                bounds=(0.9, float("inf")),
                sigmoid="linear",
                margin=1.9,
                value_at_margin=0,
            )
            stand_reward = standing * upright
            small_control = rewards.tolerance(
                actuator_forces(state, self.robot_name),
                margin=10,
                value_at_margin=0,
                sigmoid="quadratic",
            ).mean()
            small_control = (4 + small_control) / 5
            if self._move_speed == 0:
                horizontal_velocity = center_of_mass_velocity(state, self.robot_name)[[0, 1]]
                dont_move = rewards.tolerance(horizontal_velocity, margin=2).mean()
                ret_rewards.append(small_control * stand_reward * dont_move)
                ret_infos.append({
                    "small_control": small_control,
                    "stand_reward": stand_reward,
                    "dont_move": dont_move,
                    "standing": standing,
                    "upright": upright,
                })
            else:
                com_velocity = center_of_mass_velocity(state, self.robot_name)[0]
                move = rewards.tolerance(
                    com_velocity,
                    bounds=(self._move_speed, float("inf")),
                    margin=self._move_speed,
                    value_at_margin=0,
                    sigmoid="linear",
                )
                move = (5 * move + 1) / 6
                reward = small_control * stand_reward * move
                ret_rewards.append(reward)
                ret_infos.append({
                    "stand_reward": stand_reward,
                    "small_control": small_control,
                    "move": move,
                    "standing": standing,
                    "upright": upright,
                })
        ret_rewards = torch.tensor(ret_rewards)
        return ret_rewards  # , ret_infos



import math

from metasim.cfg.checkers import AndOp, DetectedChecker, RelativeBboxDetector, RotationShiftChecker
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, TaskType
from metasim.utils import configclass


@configclass
class RotatePinkBlockRightMetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.CALVIN
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    checker = AndOp([
        RotationShiftChecker(
            obj_name="block_pink",
            radian_threshold=-60 / 180 * math.pi,
        ),
        DetectedChecker(
            obj_name="block_pink",
            detector=RelativeBboxDetector(
                base_obj_name="block_pink",
                relative_pos=(0.0, 0.0, 0.0),
                relative_quat=(1.0, 0.0, 0.0, 0.0),
                checker_lower=(-0.1, -0.1, -0.1),
                checker_upper=(0.1, 0.1, 0.1),
                # debug_vis=True,
            ),
        ),
    ])



from __future__ import annotations

import math

from metasim.utils import configclass

from .base_robot_metacfg import BaseActuatorMetaCfg, BaseRobotMetaCfg


@configclass
class AntMetaCfg(BaseRobotMetaCfg):
    name: str = "ant"
    num_joints: int = 8
    freejoint: bool = True
    filepath: str = "roboverse_data/robots/ant/usd/ant.usd"
    mjcf_path: str = "roboverse_data/robots/ant/mjcf/ant.xml"
    urdf_path: str = "roboverse_data/robots/ant/urdf/ant.urdf"
    enabled_gravity: bool = True
    enabled_self_collisions: bool = True

    # Define actuators for each joint - control range from -1.0 to 1.0 with gear 15
    actuators: dict[str, BaseActuatorMetaCfg] = {
        "hip_1": BaseActuatorMetaCfg(velocity_limit=30.0),
        "ankle_1": BaseActuatorMetaCfg(velocity_limit=30.0),
        "hip_2": BaseActuatorMetaCfg(velocity_limit=30.0),
        "ankle_2": BaseActuatorMetaCfg(velocity_limit=30.0),
        "hip_3": BaseActuatorMetaCfg(velocity_limit=30.0),
        "ankle_3": BaseActuatorMetaCfg(velocity_limit=30.0),
        "hip_4": BaseActuatorMetaCfg(velocity_limit=30.0),
        "ankle_4": BaseActuatorMetaCfg(velocity_limit=30.0),
    }

    joint_limits: dict[str, tuple[float, float]] = {
        "hip_1": (-40 * math.pi / 180, 40 * math.pi / 180),  # -40 to 40 degrees
        "ankle_1": (30 * math.pi / 180, 100 * math.pi / 180),  # 30 to 100 degrees
        "hip_2": (-40 * math.pi / 180, 40 * math.pi / 180),  # -40 to 40 degrees
        "ankle_2": (-100 * math.pi / 180, -30 * math.pi / 180),  # -100 to -30 degrees
        "hip_3": (-40 * math.pi / 180, 40 * math.pi / 180),  # -40 to 40 degrees
        "ankle_3": (-100 * math.pi / 180, -30 * math.pi / 180),  # -100 to -30 degrees
        "hip_4": (-40 * math.pi / 180, 40 * math.pi / 180),  # -40 to 40 degrees
        "ankle_4": (30 * math.pi / 180, 100 * math.pi / 180),  # 30 to 100 degrees
    }

    default_joint_positions: dict[str, float] = {
        "hip_1": 0.0,
        "ankle_1": 1.0,
        "hip_2": 0.0,
        "ankle_2": -1.0,
        "hip_3": 0.0,
        "ankle_3": -1.0,
        "hip_4": 0.0,
        "ankle_4": 1.0,
    }

    control_frequency_inv: int = 1  # Control frequency divider
    motor_strength: float = 15.0  # Motor gear value from XML
    dof_damping: float = 0.1  # From default joint damping in XML
    dof_friction: float = 0.0  # Not explicitly defined in XML
    dof_armature: float = 0.01  # From default joint armature in XML

    observe_base_position: bool = True
    observe_base_velocity: bool = True
    observe_joint_positions: bool = True
    observe_joint_velocities: bool = True



import torch

_METASIM_BODY_PREFIX = "metasim_body_"
_METASIM_SITE_PREFIX = "metasim_site_"
# robot_name = "h1"
# robot_name = "h1_simple_hand"


def torso_upright(envstate, robot_name: str):
    """Returns projection from z-axes of torso to the z-axes of world."""
    # env.handler.physics.named.data.xmat[f"{robot_name}/torso_link", "zz"]
    xmat = quaternion_to_matrix(envstate[f"{_METASIM_BODY_PREFIX}{robot_name}/torso_link"]["rot"])
    return xmat[2, 2]


def head_height(envstate, robot_name: str):
    """Returns the height of the torso."""
    # env.handler.physics.named.data.site_xpos[f"{robot_name}/head", "z"]
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/head"]["pos"][2]


def left_foot_height(envstate, robot_name: str):
    """Returns the height of the left foot."""
    # env.handler.physics.named.data.site_xpos[f"{robot_name}/left_foot", "z"]
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/left_foot"]["pos"][2]


def right_foot_height(envstate, robot_name: str):
    """Returns the height of the right foot."""
    # env.handler.physics.named.data.site_xpos[f"{robot_name}/right_foot", "z"]
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/right_foot"]["pos"][2]


def center_of_mass_position(envstate, robot_name: str):
    """Returns position of the center-of-mass."""
    # env.handler.physics.named.data.subtree_com[f"{robot_name}/pelvis"].copy()
    return envstate[f"{_METASIM_BODY_PREFIX}{robot_name}/pelvis"]["com"]


def center_of_mass_velocity(envstate, robot_name: str):
    """Returns the velocity of the center-of-mass."""
    # env.handler.physics.named.data.sensordata[f"{robot_name}/pelvis_subtreelinvel"].copy()
    return envstate[f"{_METASIM_BODY_PREFIX}{robot_name}/pelvis"]["com_vel"]


def body_velocity(envstate, robot_name: str):
    """Returns the velocity of the torso in local frame."""
    # env.handler.physics.named.data.sensordata[f"{robot_name}/body_velocimeter"].copy()
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/com"][f"{robot_name}/body_velocimeter"]


def torso_vertical_orientation(envstate, robot_name: str):
    """Returns the z-projection of the torso orientation matrix."""
    # env.handler.physics.named.data.xmat[f"{robot_name}/torso_link", ["zx", "zy", "zz"]]
    xmat = quaternion_to_matrix(envstate[f"{_METASIM_BODY_PREFIX}{robot_name}/torso_link"]["rot"])
    return xmat[2, :]


@DeprecationWarning
def joint_angles(envstate, robot_name: str):
    """Returns the state without global orientation or position."""
    # Skip the 7 DoFs of the free root joint.
    # env.data.qpos[7 : self.dof].copy()
    return "joint_angles not implemented. Used for perception in humanoid_bench."


@DeprecationWarning
def joint_velocities(envstate, robot_name: str):
    """Returns the joint velocities."""
    # env.data.qvel[6 : self.dof - 1].copy()
    return "joint_velocities not implemented. Used for perception in humanoid_bench."


@DeprecationWarning
def control(envstate, robot_name: str):
    """Returns a copy of the control signals for the actuators."""
    # env.data.ctrl.copy()
    return "control not implemented. Used for perception in humanoid_bench."


def actuator_forces(envstate, robot_name: str):
    """Returns a copy of the forces applied by the actuators."""
    # env.handler.data.actuator_force.copy()
    return torch.tensor([
        x for x in envstate[f"{robot_name}"]["dof_torque"].values()
    ])  # envstate[f"{robot_name}"]["dof_torque"]


def left_hand_position(envstate, robot_name: str):
    # env.handler.physics.named.data.site_xpos[f"{robot_name}/left_hand"]
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/left_hand"]["pos"]


def left_hand_velocity(envstate, robot_name: str):
    # env.handler.physics.named.data.sensordata[f"{robot_name}/left_hand_subtreelinvel"].copy()
    return envstate[f"{_METASIM_BODY_PREFIX}{robot_name}/left_hand"][f"{robot_name}/left_hand_subtreelinvel"]


def left_hand_orientation(envstate, robot_name: str):
    # env.handler.physics.named.data.site_xmat[f"{robot_name}/left_hand"]
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/left_hand"]["rot"]


def right_hand_position(envstate, robot_name: str):
    # env.handler.physics.named.data.site_xpos[f"{robot_name}/right_hand"]
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/right_hand"]["pos"]


def right_hand_velocity(envstate, robot_name: str):
    # env.handler.physics.named.data.sensordata["right_hand_subtreelinvel"].copy()
    return envstate[f"{_METASIM_BODY_PREFIX}{robot_name}/right_hand"][f"{robot_name}/right_hand_subtreelinvel"]


def right_hand_orientation(envstate, robot_name: str):
    # env.handler.physics.named.data.site_xmat["right_hand"]
    return envstate[f"{_METASIM_SITE_PREFIX}{robot_name}/right_hand"]["rot"]


#######################################################
## Extra functions
#######################################################


def z_height(envstate, robot_name: str):
    return envstate[f"{robot_name}"]["pos"][2]


#######################################################
## Helper functions
#######################################################


def quaternion_to_matrix(quat: torch.Tensor):
    q_w, q_x, q_y, q_z = quat
    R = torch.tensor(
        [
            [1 - 2 * (q_y**2 + q_z**2), 2 * (q_x * q_y - q_w * q_z), 2 * (q_x * q_z + q_w * q_y)],
            [2 * (q_x * q_y + q_w * q_z), 1 - 2 * (q_x**2 + q_z**2), 2 * (q_y * q_z - q_w * q_x)],
            [2 * (q_x * q_z - q_w * q_y), 2 * (q_y * q_z + q_w * q_x), 1 - 2 * (q_x**2 + q_y**2)],
        ],
        dtype=quat.dtype,
    )
    return R



from __future__ import annotations

import math

from metasim.utils import configclass

from .base_robot_metacfg import BaseActuatorMetaCfg, BaseRobotMetaCfg


@configclass
class WalkerMetaCfg(BaseRobotMetaCfg):
    name: str = "walker"
    num_joints: int = 6
    fix_base_link: bool = False
    filepath: str = "roboverse_data/robots/walker/usd/walker.usd"
    mjcf_path: str = "roboverse_data/robots/walker/mjcf/walker.xml"
    urdf_path: str = "roboverse_data/robots/walker/urdf/walker.urdf"
    enabled_gravity: bool = True
    enabled_self_collisions: bool = True

    # Define actuators for each joint - control range from -1.0 to 1.0 with gear values from XML
    actuators: dict[str, BaseActuatorMetaCfg] = {
        "right_hip": BaseActuatorMetaCfg(velocity_limit=30.0),
        "right_knee": BaseActuatorMetaCfg(velocity_limit=30.0),
        "right_ankle": BaseActuatorMetaCfg(velocity_limit=30.0),
        "left_hip": BaseActuatorMetaCfg(velocity_limit=30.0),
        "left_knee": BaseActuatorMetaCfg(velocity_limit=30.0),
        "left_ankle": BaseActuatorMetaCfg(velocity_limit=30.0),
    }

    # Joint limits from the XML (converted from degrees to radians)
    joint_limits: dict[str, tuple[float, float]] = {
        "right_hip": (-20 * math.pi / 180, 100 * math.pi / 180),  # -20 to 100 degrees
        "right_knee": (-150 * math.pi / 180, 0 * math.pi / 180),  # -150 to 0 degrees
        "right_ankle": (-45 * math.pi / 180, 45 * math.pi / 180),  # -45 to 45 degrees
        "left_hip": (-20 * math.pi / 180, 100 * math.pi / 180),  # -20 to 100 degrees
        "left_knee": (-150 * math.pi / 180, 0 * math.pi / 180),  # -150 to 0 degrees
        "left_ankle": (-45 * math.pi / 180, 45 * math.pi / 180),  # -45 to 45 degrees
    }

    # Default starting pose - using zeros (no explicit init_qpos in the XML)
    default_joint_positions: dict[str, float] = {
        "right_hip": 0.0,
        "right_knee": 0.0,
        "right_ankle": 0.0,
        "left_hip": 0.0,
        "left_knee": 0.0,
        "left_ankle": 0.0,
    }

    default_position: tuple[float, float, float] = (0.0, 0.0, 1.3)

    # Control constants from the XML
    control_frequency_inv: int = 1  # Control frequency divider
    motor_strength: dict[str, float] = {
        "right_hip": 100.0,  # From gear="100" in XML
        "right_knee": 50.0,  # From gear="50" in XML
        "right_ankle": 20.0,  # From gear="20" in XML
        "left_hip": 100.0,  # From gear="100" in XML
        "left_knee": 50.0,  # From gear="50" in XML
        "left_ankle": 20.0,  # From gear="20" in XML
    }

    # From default joint damping in XML
    dof_damping: float = 0.1
    # From default joint friction in XML
    dof_friction: float = 0.0
    # From default joint armature in XML
    dof_armature: float = 0.01

    # Observation space configuration
    observe_base_position: bool = True
    observe_base_velocity: bool = True
    observe_joint_positions: bool = True
    observe_joint_velocities: bool = True



from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PushButtonsMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 500
    traj_filepath = "roboverse_data/trajs/rlbench/push_buttons/v2"
    objects = [
        ArticulationObjMetaCfg(
            name="push_buttons_target0",
            filepath="roboverse_data/assets/rlbench/push_buttons/push_buttons_target0/usd/push_buttons_target0.usd",
        ),
        ArticulationObjMetaCfg(
            name="push_buttons_target1",
            filepath="roboverse_data/assets/rlbench/push_buttons/push_buttons_target1/usd/push_buttons_target1.usd",
        ),
        ArticulationObjMetaCfg(
            name="push_buttons_target2",
            filepath="roboverse_data/assets/rlbench/push_buttons/push_buttons_target2/usd/push_buttons_target2.usd",
        ),
    ]
    # TODO: add checker



from __future__ import annotations

import numpy as np
import torch
from loguru import logger as log
from rlbench.backend.robot import Robot

from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.types import Action, EnvState, Extra, Obs, Reward, Success, TimeOut
from metasim.utils import to_snake_case
from pyrep import PyRep
from pyrep.objects.object import Object
from pyrep.robots.arms.panda import Panda
from pyrep.robots.end_effectors.panda_gripper import PandaGripper

from ..base import BaseSimHandler

# TODO: try best to be independent from RLBench

## Joint names correspondence dict
jd = {
    "franka": {
        "Panda_joint1": "panda_joint1",
        "Panda_joint2": "panda_joint2",
        "Panda_joint3": "panda_joint3",
        "Panda_joint4": "panda_joint4",
        "Panda_joint5": "panda_joint5",
        "Panda_joint6": "panda_joint6",
        "Panda_joint7": "panda_joint7",
        "Panda_gripper_joint1": "panda_finger_joint1",
        "Panda_gripper_joint2": "panda_finger_joint2",
    }
}


class PyrepHandler(BaseSimHandler):
    def launch(self) -> None:
        self.sim = PyRep()
        self.sim.launch("third_party/rlbench/rlbench/task_design.ttt", headless=False)  # TODO: set headless from cfg
        # self.sim.set_simulation_timestep(1 / 60)  # Control frequency 60Hz, TODO: set control frequency from cfg
        task_name = to_snake_case(self.task.__class__.__name__.replace("MetaCfg", ""))
        base_object = self.sim.import_model(f"third_party/rlbench/rlbench/task_ttms/{task_name}.ttm")

        self.arm, self.gripper = Panda(), PandaGripper()
        self.robot_inst = Robot(self.arm, self.gripper)
        self.sim.start()
        self.sim.step()
        # TODO: initialize rlbench scene and task

    ############################################################
    ## Gymnasium main methods
    ############################################################
    def step(self, action: list[Action]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
        action = action[0]["dof_pos_target"]
        arm_action = [action[jd[self.robot.name][joint.get_name()]] for joint in self.arm.joints]
        gripper_action = [action[jd[self.robot.name][joint.get_name()]] for joint in self.gripper.joints]
        self.arm.set_joint_target_positions(arm_action)
        self.gripper.set_joint_target_positions(gripper_action)
        self.sim.step()
        self.arm.set_joint_target_positions(self.arm.get_joint_positions())
        self.gripper.set_joint_target_positions(self.gripper.get_joint_positions())
        return None, None, torch.tensor([False]), torch.tensor([False]), None

    def reset(self) -> tuple[Obs, Extra]:
        return None, None

    def render(self) -> None:
        pass

    def close(self) -> None:
        self.sim.stop()
        self.sim.shutdown()

    ############################################################
    ## Set states
    ############################################################
    def set_states(self, states: list[EnvState]) -> None:
        assert len(states) == 1  # PyRep only supports one env
        state = states[0]

        ## Set robot
        joint_pos = np.zeros(len(self.arm.joints))
        for i, joint_inst in enumerate(self.arm.joints):
            joint_name = joint_inst.get_name()
            joint_pos[i] = state[self.robot.name]["dof_pos"][joint_name]
        self.arm.set_joint_positions(joint_pos[:7])
        self.gripper.set_joint_positions(joint_pos[7:])
        # TODO: set robot pose #!

        ## Set objects
        for obj in self.objects:
            obj_inst = Object.get_object(obj.name)
            obj_inst.set_position(state[obj.name]["pos"].numpy())
            obj_inst.set_orientation(state[obj.name]["rot"].numpy())

            log.debug("Set object", obj.name, "to", state[obj.name]["pos"], state[obj.name]["rot"])

    def set_pose(self, obj_name: str, pos: torch.Tensor, rot: torch.Tensor) -> None:
        raise NotImplementedError

    ############################################################
    ## Get states
    ############################################################
    def get_states(self) -> list[EnvState]:
        pass

    ############################################################
    ## Utils
    ############################################################
    def is_running(self) -> bool:
        return True

    def get_object_joint_names(self, object: ArticulationObjMetaCfg) -> list[str]:
        assert isinstance(object, ArticulationObjMetaCfg)
        pass

    def get_robot_joint_limits(self) -> torch.Tensor:
        return True



from metasim.cfg.objects import ArticulationObjMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class ChangeChannelMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/change_channel/v2"
    objects = [
        ArticulationObjMetaCfg(
            name="tv_remote",
            filepath="roboverse_data/assets/rlbench/change_channel/tv_remote/usd/tv_remote.usd",
        ),
        RigidObjMetaCfg(
            name="tv_frame",
            filepath="roboverse_data/assets/rlbench/change_channel/tv_frame/usd/tv_frame.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PlaceCupsMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/place_cups/v2"
    objects = [
        RigidObjMetaCfg(
            name="place_cups_holder_base",
            filepath="roboverse_data/assets/rlbench/place_cups/place_cups_holder_base/usd/place_cups_holder_base.usd",
            physics=PhysicStateType.XFORM,
        ),
        RigidObjMetaCfg(
            name="mug_visual0",
            filepath="roboverse_data/assets/rlbench/place_cups/mug_visual1/usd/mug_visual1.usd",  # reuse same asset
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="mug_visual1",
            filepath="roboverse_data/assets/rlbench/place_cups/mug_visual1/usd/mug_visual1.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="mug_visual2",
            filepath="roboverse_data/assets/rlbench/place_cups/mug_visual1/usd/mug_visual1.usd",  # reuse same asset
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="mug_visual3",
            filepath="roboverse_data/assets/rlbench/place_cups/mug_visual1/usd/mug_visual1.usd",  # reuse same asset
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from metasim.cfg.checkers import WalkChecker
from metasim.utils import configclass

from .base_locomotion_env import BaseLocomotionReward, HumanoidTaskMetaCfg


class WalkReward(BaseLocomotionReward):
    _move_speed = 1
    success_bar = 700


@configclass
class WalkMetaCfg(HumanoidTaskMetaCfg):
    episode_length = 1000
    # traj_filepath = "roboverse_data/trajs/humanoidbench/walk/v2/h1_v2.pkl"
    traj_filepath = "roboverse_data/trajs/humanoidbench/walk/v2/initial_state_v2.json"
    checker = WalkChecker()
    reward_functions = [WalkReward]
    reward_weights = [1.0]



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PutBooksOnBookshelfMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/put_books_on_bookshelf/v2"
    objects = [
        RigidObjMetaCfg(
            name="bookshelf_visual",
            filepath="roboverse_data/assets/rlbench/put_books_on_bookshelf/bookshelf_visual/usd/bookshelf_visual.usd",
            physics=PhysicStateType.GEOM,
        ),
        RigidObjMetaCfg(
            name="book0_visual",
            filepath="roboverse_data/assets/rlbench/put_books_on_bookshelf/book0_visual/usd/book0_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="book1_visual",
            filepath="roboverse_data/assets/rlbench/put_books_on_bookshelf/book1_visual/usd/book1_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="book2_visual",
            filepath="roboverse_data/assets/rlbench/put_books_on_bookshelf/book2_visual/usd/book2_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from metasim.cfg.checkers import DetectedChecker, RelativeBboxDetector
from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class HockeyMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 300
    traj_filepath = "roboverse_data/trajs/rlbench/hockey/v2"
    objects = [
        RigidObjMetaCfg(
            name="hockey_goal_visual",
            filepath="roboverse_data/assets/rlbench/hockey/hockey_goal_visual/usd/hockey_goal_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="hockey_stick",
            filepath="roboverse_data/assets/rlbench/hockey/hockey_stick/usd/hockey_stick.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="hockey_ball",
            filepath="roboverse_data/assets/rlbench/hockey/hockey_ball/usd/hockey_ball.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    checker = DetectedChecker(
        obj_name="hockey_ball",
        detector=RelativeBboxDetector(
            base_obj_name="hockey_goal_visual",
            relative_pos=[0, -0.022753269108366, 0.083899194072109],
            relative_quat=[0.00012045016205838, 0.99993911881697, 0.010847761490316, 0.0020174791069082],
            checker_lower=[-0.056, -0.028, 0],
            # checker_upper=[0.056, 0.028, 0.168],
            checker_upper=[0.056, 0.045, 0.168],  # XXX
            # debug_vis=True,
        ),
    )



import unittest

from metasim.cfg.cameras import PinholeCameraMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.constants import SimType
from metasim.scripts.replay_demo import get_states
from metasim.utils.demo_util import get_traj
from metasim.utils.setup_util import get_sim_env_class


def do_all_tests(task, simulator, robot):
    camera = PinholeCameraMetaCfg(
        name="camera",
        data_types=["rgb", "depth"],
        width=256,
        height=256,
        pos=(1.5, -1.5, 1.5),  # near
        # pos=(12.0, -12.0, 20.0),  # far
        look_at=(0.0, 0.0, 0.0),
    )
    scenario = ScenarioMetaCfg(
        task=task,
        robot=robot,
        cameras=[camera],
        sim=simulator,
    )
    env_class = get_sim_env_class(SimType(scenario.sim))
    env = env_class(scenario, 1)

    init_states, all_actions, all_states = get_traj(scenario.task, scenario.robot, env.handler)

    for i in range(100):
        states = get_states(all_states, i, 1)
        env.handler.set_states(states)
        env.handler.refresh_render()

    env.handler.close()


# Test case to detect runtime errors
class SimulatorRuntimeErrors(unittest.TestCase):
    testing_task = "close_box"
    testing_robot = "franka"

    def test_pybullet(self):
        do_all_tests(self.testing_task, "pybullet", self.testing_robot)

    def test_mujoco(self):
        do_all_tests(self.testing_task, "mujoco", self.testing_robot)


class TaskRuntimeErrors(unittest.TestCase):
    def test_rlbench(self):
        do_all_tests("close_box", "pybullet", "franka")

    def test_basic(self):
        do_all_tests("pick_cube", "pybullet", "franka")


unittest.main()



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, PhysicStateType, TaskType
from metasim.utils import configclass


@configclass
class OpensdorRotBladesRightScissors593MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/037_scissors_google_16k/037_scissors_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/037_blades_right/20240826-215126_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors594MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/037_scissors_google_16k/037_scissors_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/037_blades_right/20240826-214018_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors595MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/037_scissors_google_16k/037_scissors_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/037_blades_right/20240826-213331_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors596MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/037_scissors_google_16k/037_scissors_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/037_blades_right/20240826-220102_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors597MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/037_scissors_google_16k/037_scissors_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/037_blades_right/20240826-214010_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors598MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/ycb_16k_backup/037_scissors_google_16k/037_scissors_google_16k.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/037_blades_right/20240826-220124_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors1341MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/2b1120c4b517409184413ed709c57a66/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/2b1120c4b517409184413ed709c57a66_blades_right/20240826-212108_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors1342MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/2b1120c4b517409184413ed709c57a66/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/2b1120c4b517409184413ed709c57a66_blades_right/20240826-213337_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors1343MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/2b1120c4b517409184413ed709c57a66/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/2b1120c4b517409184413ed709c57a66_blades_right/20240826-213116_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors1344MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/2b1120c4b517409184413ed709c57a66/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/2b1120c4b517409184413ed709c57a66_blades_right/20240826-212815_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors1345MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/2b1120c4b517409184413ed709c57a66/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/2b1120c4b517409184413ed709c57a66_blades_right/20240826-211923_no_interaction/trajectory-unified_wo_traj_v2.pkl"


@configclass
class OpensdorRotBladesRightScissors1346MetaCfg(BaseTaskMetaCfg):
    source_benchmark = BenchmarkType.OPEN6DOR
    task_type = TaskType.TABLETOP_MANIPULATION
    episode_length = 600
    objects = [
        PrimitiveCubeMetaCfg(
            name="table",
            mass=1,
            color=[255, 255, 255],
            size=[0.6, 0.8, 0.3],
        ),
        RigidObjMetaCfg(
            name="scissors",
            urdf_path=(
                "data_isaaclab/assets/open6dor/objaverse_rescale/2b1120c4b517409184413ed709c57a66/material_2.urdf"
            ),
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    traj_filepath = "data_isaaclab/source_data/open6dor/task_refine_rot_only/rot_ins/2b1120c4b517409184413ed709c57a66_blades_right/20240826-212634_no_interaction/trajectory-unified_wo_traj_v2.pkl"
